#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import codecs
import json
from datetime import timedelta, datetime

from airflow.models import DAG
from airflow.contrib.operators.gcs_to_bq import GoogleCloudStorageToBigQueryOperator
from airflow.contrib.sensors.gcs_sensor import GoogleCloudStoragePrefixSensor
from airflow.contrib.operators.bigquery_operator import BigQueryOperator
from airflow.operators.dummy_operator import DummyOperator

from airflow.hooks.base_hook import BaseHook
from airflow.contrib.operators.slack_webhook_operator import SlackWebhookOperator

import pendulum

local_tz = pendulum.timezone("Asia/Seoul")

# 에러 발생 시에 Slack으로 통지하기 위한 function을 정의한다.
# slack aleat 스크립트는 Kaxil에게 받은 것이다.
# 당사자 허가를 받아 게재하였으며 다음 기사에 공개되어 있다.
# https://medium.com/datareply/integrating-slack-alerts-in-airflow-c9dcd155105
def task_fail_slack_alert(context):
    slack_webhook_token = BaseHook.get_connection('slack').password
    slack_msg = """
            :red_circle: Task Failed.
            *Task*: {task}
            *Dag*: {dag}
            *Execution Time*: {exec_date}
            *Log Url*: {log_url}
            """.format(
        task=context.get('task_instance').task_id,
        dag=context.get('task_instance').dag_id,
        ti=context.get('task_instance'),
        exec_date=local_tz.convert(context.get('execution_date')),
        log_url=context.get('task_instance').log_url
    )
    failed_alert = SlackWebhookOperator(
        task_id='slack_test',
        http_conn_id='slack',
        webhook_token=slack_webhook_token,
        message=slack_msg,
        username='airflow',
        dag=dag)
    return failed_alert.execute(context=context)


default_args = {
    'owner': 'Airflow',
    'start_date': datetime(2019, 9, 30, tzinfo=local_tz),
    # 'end_date': datetime(2020, 12, 31), # 필요가 있으면 설정하기
    'depends_on_past': True,
    'retries': 1,
    'retry_delay': timedelta(minutes=1),
    # 'on_failure_callback': task_fail_slack_alert,
    'catchup_by_default': True
}

dag = DAG('dwh_samle', schedule_interval='@daily',
          catchup=False, default_args=default_args)

PROJECT_NAME = "shuwa-gcp-book"
BUCKET_NAME = "composer_source"


"""
참고로 이 예제에서는 json 및 sql 파일에 의해 워크플로에서 처리하는 테이블의 종류나 Query를 외부 파일로 정의하고 그것을 읽어 들이는 방법으로 처리하고 있다.
"""

# 복수 테이블 JOIN 시에 이용하는 더미 오퍼레이터를 먼저 정의한다.
wait_sales = DummyOperator(
    task_id='wait_sales',
    trigger_rule='all_success',
    dag=dag,
)
wait_stock = DummyOperator(
    task_id='wait_stock',
    trigger_rule='all_success',
    dag=dag,
)
wait_customer = DummyOperator(
    task_id='wait_customer',
    trigger_rule='all_success',
    dag=dag,
)

# import로 load；yyyymmdd가 있는 일별 테이블로 한다.
with codecs.open('/home/airflow/gcs/dags/tables/tables.json', 'r', 'utf-8') as f:
    tables = json.load(f)

    for i in tables:
        # GCS 버킷 파일을 감지한다.
        sense_file = GoogleCloudStoragePrefixSensor(
            task_id='sense_file_{0}'.format(i['name']),
            bucket=BUCKET_NAME,
            prefix='{0}_{1}'.format(
                i['name'], '{{(execution_date + macros.timedelta(days=2)).strftime("%Y%m%d")}}'),
            dag=dag
        )
        # 파일을 import로 load하고 일별 테이블로 한다.
        # ‘source_objects’ 파라미터 부문에서 *(와일드카드)를 사용하고 있다.
        # 이것은 연월일 뒤에 시각 정보가 포함되는 경우가 있어 이에 대한 대응안이다.
        with codecs.open('/home/airflow/gcs/dags/schema/{0}.json'.format(i['name']), 'r', 'utf-8') as f:
            schema = json.load(f)
            gcs_to_bq = GoogleCloudStorageToBigQueryOperator(
                task_id='gcs_to_bq_{0}'.format(i['name']),
                bucket=BUCKET_NAME,
                source_objects=['{0}_{1}*.csv'.format(
                    i['name'], '{{(execution_date + macros.timedelta(days=2)).strftime("%Y%m%d")}}')],
                source_format='CSV',
                schema_fields=schema,
                skip_leading_rows=1,
                destination_project_dataset_table='{0}.import.{1}_{2}'.format(
                    PROJECT_NAME, i['name'], '{{(execution_date + macros.timedelta(days=2)).strftime("%Y%m%d")}}'),
                create_disposition='CREATE_IF_NEEDED',
                write_disposition='WRITE_TRUNCATE',
                dag=dag,
            )
        # import 테이블에 클렌징용 Query를 실행하고 결과를 cold 테이블로 저장한다.
        # 일별 테이블로 한다.
        with codecs.open('/home/airflow/gcs/dags/query/to_cold_{0}.sql'.format(i['name']), 'r', 'utf-8') as f:
            query = f.read()
            query = query.format(
                '{{(execution_date + macros.timedelta(days=2)).strftime("%Y%m%d")}}', 'import')
            import_to_cold = BigQueryOperator(
                task_id='import_to_cold_{0}'.format(i['name']),
                sql=query,
                destination_dataset_table='{0}.cold.{1}_{2}'.format(
                    PROJECT_NAME, i['name'], '{{(execution_date + macros.timedelta(days=2)).strftime("%Y%m%d")}}'),
                create_disposition='CREATE_IF_NEEDED',
                write_disposition='WRITE_TRUNCATE',
                use_legacy_sql=False,
                dag=dag,
            )
        # cold 테이블을 큰 work 테이블에 append한다.
        # 이력형 테이블로 하는 경우 fromDate, toDate를 부여해 Merge하는 등 Query에 대한 고려가 필요하다. 참고로 Merge할 경우는 BigQueryOperator의 파라미터가 변하는 점에 주의하기 바란다. ‘destination_dataset_table’의 지정을 할 수 없게 된다.
        with codecs.open('/home/airflow/gcs/dags/query/to_work_{0}.sql'.format(i['name']), 'r', 'utf-8') as f:
            query = f.read()
            query = query.format(
                'cold', '{{(execution_date + macros.timedelta(days=2)).strftime("%Y%m%d")}}')
            cold_to_work = BigQueryOperator(
                task_id='cold_to_work_{0}'.format(i['name']),
                sql=query,
                destination_dataset_table='{0}.work.{1}'.format(
                    PROJECT_NAME, i['name']),
                create_disposition='CREATE_IF_NEEDED',
                write_disposition='WRITE_APPEND',
                use_legacy_sql=False,
                dag=dag,
            )
        # 개별로 정의한 Operator의 의존관계를 정의한다.
        sense_file >> gcs_to_bq >> import_to_cold >> cold_to_work

        # sales 관련 테이블이 work 단계에 모두 갖추어지는 것을 더미 오퍼레이터를 사용해 실시한다.
        if i['join_sales'] == 1:
            cold_to_work >> wait_sales

        # stock 관련 테이블이 work 단계에 모두 갖추어지는 것을 더미 오퍼레이터를 사용해 실시한다.
        if i['join_stock'] == 1:
            cold_to_work >> wait_stock

        # customer 관련 테이블이 work 단계에 모두 갖추어지는 것을 더미 오퍼레이터를 사용해 실시한다.
        if i['join_customer'] == 1:
            cold_to_work >> wait_customer

# work 데이터세트에 있는 sales 관련 테이블을 결합한다.
with codecs.open('/home/airflow/gcs/dags/query/join_sales.sql', 'r', 'utf-8') as f:
    query = f.read()
    query = query.format(PROJECT_NAME)
    join_sales = BigQueryOperator(
        task_id='join_sales',
        sql=query,
        destination_dataset_table='{0}.warm.sales_joined'.format(PROJECT_NAME),
        create_disposition='CREATE_IF_NEEDED',
        write_disposition='WRITE_TRUNCATE',
        use_legacy_sql=False,
        dag=dag
    )
    wait_sales >> join_sales


# work 데이터세트에 있는 stock 관련 테이블을 통합한다.
with codecs.open('/home/airflow/gcs/dags/query/join_stock.sql', 'r', 'utf-8') as f:
    query = f.read()
    query = query.format(PROJECT_NAME)
    join_stock = BigQueryOperator(
        task_id='join_stock',
        sql=query,
        destination_dataset_table='{0}.warm.stock_joined'.format(PROJECT_NAME),
        create_disposition='CREATE_IF_NEEDED',
        write_disposition='WRITE_TRUNCATE',
        use_legacy_sql=False,
        dag=dag
    )
    wait_stock >> join_stock

# work 데이터세트에 있는 stock 관련 테이블을 결합한다.
with codecs.open('/home/airflow/gcs/dags/query/join_customer.sql', 'r', 'utf-8') as f:
    query = f.read()
    query = query.format(PROJECT_NAME)
    join_customer = BigQueryOperator(
        task_id='join_customer',
        sql=query,
        destination_dataset_table='{0}.warm.customer_joined'.format(
            PROJECT_NAME),
        create_disposition='CREATE_IF_NEEDED',
        write_disposition='WRITE_TRUNCATE',
        use_legacy_sql=False,
        dag=dag
    )
    wait_customer >> join_customer
